import time
from pathlib import Path

import evdev


class VirtualKeyboard:
    def __init__(self, keyboard_device_path: Path):
        self.device = evdev.InputDevice(keyboard_device_path)

    CHAR_MAP = {
        " ": "KEY_SPACE",
        ",": "KEY_COMMA",
        ".": "KEY_DOT",
        "-": "KEY_MINUS",
        "_": "KEY_MINUS",
        "/": "KEY_SLASH",
        ":": "KEY_SEMICOLON",
        ";": "KEY_SEMICOLON",
        "'": "KEY_APOSTROPHE",
        '"': "KEY_APOSTROPHE",
        "!": "KEY_1",
        "?": "KEY_SLASH",
        "£": "KEY_3",
    }

    def _send_key(self, key_event: str, shift: bool = False) -> None:
        if shift:
            self.device.write(evdev.ecodes.EV_KEY, evdev.ecodes.KEY_LEFTSHIFT, 1)
        self.device.write(evdev.ecodes.EV_KEY, evdev.ecodes.ecodes[key_event], 1)
        self.device.write(evdev.ecodes.EV_KEY, evdev.ecodes.ecodes[key_event], 0)
        if shift:
            self.device.write(evdev.ecodes.EV_KEY, evdev.ecodes.KEY_LEFTSHIFT, 0)

        # Process the above write events
        self.device.syn()  # type: ignore
        time.sleep(0.01)  # this helps prevent the letters from getting jumbled up

    def type_text(self, text: str) -> None:
        for char in text:
            if char.isalpha():
                key = f"KEY_{char.upper()}"
                self._send_key(key, shift=char.isupper())
            elif char.isdigit():
                self._send_key(f"KEY_{char}")
            elif char in self.CHAR_MAP:
                shift = char in '!?:"_£'
                self._send_key(self.CHAR_MAP[char], shift)



import configparser
from pathlib import Path

CONFIG_DIR = Path.home() / ".config/speech2caret"
CONFIG_FILE = CONFIG_DIR / "config.ini"


def get_config() -> configparser.ConfigParser:
    """Get user configuration.

    If the config file doesn't exist, it will be created.
    """
    CONFIG_DIR.mkdir(exist_ok=True)
    config = configparser.ConfigParser()

    # Create a default config file
    if not CONFIG_FILE.is_file():
        config["speech2caret"] = {
            "# example:\n# keyboard_device_path": "/dev/input/by-path/pci-0000:00:1.0-usb-0:1:1.0-event-kbd",
            "# start_stop_key": "KEY_F11",
            "# resume_pause_key": "KEY_F12",
            "keyboard_device_path": "",
            "start_stop_key": "",
            "resume_pause_key": "",
        }
        with open(CONFIG_FILE, "w") as f:
            f.write(
                "# This is the configuration file for speech2caret.\n"
                "# You can find an explanation of the options in the GitHub README.md: https://github.com/asmith26/speech2caret\n\n"
            )
            config.write(f)

    config.read(CONFIG_FILE)
    return config



import argparse
import asyncio
import sys
import tempfile
from pathlib import Path

from loguru import logger

from speech2caret.config import get_config


def validate_inputs(keyboard_device_path: Path, start_stop_key: str, resume_pause_key: str) -> None:
    """Validate and return the inputs for the main function."""
    config_help = "Edit the config file (see https://github.com/asmith26/speech2caret/tree/main#configuration) or pass in CLI arguments (see `speech2caret --help`)"
    if keyboard_device_path == Path("."):
        logger.error(f"keyboard_device_path not set. {config_help}.")
        sys.exit(1)
    if not keyboard_device_path.exists():
        logger.error(f"Keyboard device does not exist: {keyboard_device_path}")
        sys.exit(1)
    if not start_stop_key:
        logger.error(f"start_stop_key not set. {config_help}.")
        sys.exit(1)
    if not resume_pause_key:
        logger.error(f"resume_pause_key not set. {config_help}.")
        sys.exit(1)


async def listen_keyboard_events(
    keyboard_device_path: Path,
    start_stop_key: str,
    resume_pause_key: str,
) -> None:  # pragma: no cover
    # Put import statements here to improve CLI performance.
    import evdev

    from speech2caret.recorder import Recorder
    from speech2caret.speech_to_text import SpeechToText
    from speech2caret.virtual_keyboard import VirtualKeyboard

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_audio_fp = Path(temp_dir) / "speech2caret.wav"
        recorder = Recorder(tmp_audio_fp)
        stt: SpeechToText = SpeechToText()
        vkeyboard = VirtualKeyboard(keyboard_device_path)

        logger.info(f"Listening on {keyboard_device_path}")
        logger.info(f"Start/Stop: {start_stop_key}")
        logger.info(f"Resume/Pause: {resume_pause_key}")
        logger.info(f"Temporary audio file: {tmp_audio_fp}")

        try:
            async for event in vkeyboard.device.async_read_loop():
                if event.type == evdev.ecodes.EV_KEY:  # if input event is a keyboard key event
                    key_event: evdev.KeyEvent = evdev.categorize(event)  # type: ignore
                    if key_event.keystate == evdev.events.KeyEvent.key_down:
                        if key_event.keycode == start_stop_key:
                            if not recorder.is_recording and not recorder.is_paused:
                                logger.info("=== Start recording ===")
                                asyncio.create_task(recorder.start_recording())
                            else:
                                logger.info("Stopping recording...")
                                recorder.save_recording()
                                text = stt.transcribe(recorder.audio_fp)
                                logger.info(f"Transcribed text: {text}")
                                vkeyboard.type_text(text)
                                recorder.delete_audio_file()

                        elif key_event.keycode == resume_pause_key:
                            if not recorder.is_recording and recorder.is_paused:
                                logger.info("Resuming recording...")
                                asyncio.create_task(recorder.start_recording(is_resume=True))
                            elif recorder.is_recording and not recorder.is_paused:
                                logger.info("Pausing recording...")
                                recorder.pause_recording()
                            elif not recorder.is_recording and not recorder.is_paused:
                                logger.warning("You must start recording before resume/pause")

        except Exception:
            logger.exception("An unexpected error occurred in the event loop.")


def main() -> None:  # pragma: no cover
    """Use your speech to write the current caret position!"""
    config = get_config()
    parser = argparse.ArgumentParser(description="Use your speech to write to the current caret position.")
    parser.add_argument(
        "--keyboard-device-path",
        type=Path,
        default=config.get("speech2caret", "keyboard_device_path", fallback=None),
        help="Path to the keyboard device.",
    )
    parser.add_argument(
        "--start-stop-key",
        type=str,
        default=config.get("speech2caret", "start_stop_key", fallback=None),
        help="Key to start/stop recording.",
    )
    parser.add_argument(
        "--resume-pause-key",
        type=str,
        default=config.get("speech2caret", "resume_pause_key", fallback=None),
        help="Key to resume/pause recording.",
    )
    args = parser.parse_args()

    validate_inputs(args.keyboard_device_path, args.start_stop_key, args.resume_pause_key)
    try:
        asyncio.run(
            listen_keyboard_events(
                args.keyboard_device_path,
                args.start_stop_key,
                args.resume_pause_key,
            )
        )
    except KeyboardInterrupt:
        logger.info("Successfully exited speech2caret. Goodbye!")
        sys.exit(0)


if __name__ == "__main__":  # pragma: no cover
    main()



import asyncio
import wave
from pathlib import Path
from typing import List

import numpy as np
import sounddevice
from _sounddevice import ffi
from loguru import logger


class Recorder:
    def __init__(self, audio_fp: Path):
        self.audio_fp = audio_fp
        self.channels = 2  # stereo
        self.sampwidth = 2  # 2 bytes (16-bit audio)
        self.sample_rate = 44100
        self.audio_format = "int16"
        self.audio_data: List[np.ndarray] = []
        self.is_recording = False
        self.is_paused = False
        # Use an asyncio.Event to signal when recording should stop
        self._stop_event = asyncio.Event()

    def delete_audio_file(self) -> None:
        logger.info(f"Deleting audio file: {self.audio_fp}")
        self.audio_fp.unlink(missing_ok=True)

    def _callback(self, indata: np.ndarray, frames: int, time: ffi.CData, status: sounddevice.CallbackFlags) -> None:
        if status:
            logger.debug(status)
        self.audio_data.append(indata.copy())

    async def start_recording(self, is_resume: bool = False) -> None:
        self.is_recording = True
        self.is_paused = False
        # Clear the stop event flag, allowing the recording loop to run.
        self._stop_event.clear()

        # If we're starting a new recording, clear the old audio data.
        if not is_resume:
            self.audio_data = []

        with sounddevice.InputStream(
            samplerate=self.sample_rate, channels=self.channels, dtype=self.audio_format, callback=self._callback
        ):
            # Wait efficiently until the event is set to True with _stop_event.set() (instead of polling in a loop with `await asyncio.sleep(0.1)`)
            await self._stop_event.wait()

    def pause_recording(self) -> None:
        self.is_recording = False
        self.is_paused = True
        # Signal the event to stop the recording loop.
        self._stop_event.set()

    def save_recording(self) -> None:
        self.is_recording = False
        self.is_paused = False
        # Signal the event to stop the recording loop.
        self._stop_event.set()

        # Convert the list to a numpy array
        audio_data = np.concatenate(self.audio_data, axis=0)

        # Save the recorded data as a WAV file
        with wave.open(str(self.audio_fp), "wb") as wf:
            wf.setnchannels(self.channels)
            wf.setsampwidth(self.sampwidth)
            wf.setframerate(self.sample_rate)
            wf.writeframes(audio_data.tobytes())



from pathlib import Path

import torch
from transformers import pipeline  # type: ignore


class SpeechToText:
    def __init__(self, model_name: str = "openai/whisper-base.en") -> None:
        self.pipe = pipeline(
            "automatic-speech-recognition",
            model=model_name,
            # assistant_model="openai/whisper-tiny.en",
            device="cpu",
            torch_dtype=torch.float32,
            chunk_length_s=30,  # split audio into 30 s pieces (allowing for longer audio)
            ignore_warning=True,  # ignore warnings about chunk length todo explore this further
        )

    def transcribe(self, audio_fp: Path) -> str:
        result = self.pipe(str(audio_fp), batch_size=1)
        return result["text"].strip()  # type: ignore
